// NFC [NHN Foundation Class]
//
//
// Copyright (C) 2006, nhn corp
// All rights reserved.

// $Id: IkXml.cpp,v 1.2 2006/07/29 06:10:00 taekwonv Exp $

#include "stdafx.h"
#include "IkXml.h"

using namespace std;
using namespace nfc;



#ifndef ASSERT
	#define ASSERT(s)       (void(0))
#endif // ASSERT
#ifndef TRACE
	#define TRACE(s)        (void(0))
#endif // TRACE
#ifndef ABORT
	#define ABORT(s)        (void(0))
#endif // ABORT



#define XMLENCODE_BUFFSIZE	128


#define STR_APPEND(strRet, pch, len) \
	strRet.append((pch), (len))

#define STR_ASSIGN(strRet, pch, len) \
	strRet.assign((pch), (len))

#define STR_ISEMPTY(str) str.empty()

#define STR_LENGTH(str) str.length()

#define STR_C_STR(str) str.c_str()

inline void STR_RTRIM(string & str)
{
	if (str.empty())
		return;

	string::iterator it = str.end();
	do
	{
		it--;
	}
	while (it >= str.begin() && IS_WHITESPACE(*it));

	it++;

	if (it < str.end())
		str.resize( it - str.begin() );
}

string nfc::get_str_tok(const char * szSrc, const char * szSplit, int nIndex)
{
	if (szSrc == NULL || strlen(szSrc) == 0)
		return "";

	string strRet;
	const int nSplitLen = strlen(szSplit);
	const char * pTokStart = szSrc;
	for (int i=0 ; *pTokStart != '\0' ; i++)
	{
		const char * pTokEnd = strstr(pTokStart, szSplit);
		if (pTokEnd == NULL)
		{
			if ( i == nIndex )
			{
				strRet = pTokStart;
			}

			break;
		}
		else
		{
			if ( i == nIndex )
			{
				STR_ASSIGN(strRet, pTokStart, pTokEnd-pTokStart);
				break;
			}

			pTokStart = pTokEnd + nSplitLen;
		}
	}

	return strRet;
}

string CIkXMLDom::EncodeXML(const char * szSrc)
{
	if (szSrc == NULL)
		return "";

	string strRet;

	char pszEncodedText[XMLENCODE_BUFFSIZE + 1];

	const char *pch = szSrc;
	
	int i = 0;
	while (*pch != '\0')
	{
		if ( *pch == '&' || *pch == '<' || *pch == '>' || *pch == '\'' || *pch == '"' )
		{
			if (i > 0)
			{
				pszEncodedText[i] = 0;
				strRet += pszEncodedText;
				i = 0;
			}
			
			switch (*pch)
			{
			case '&':
				strRet += "&amp;";
				break;
			case '<':
				strRet += "&lt;";
				break;
			case '>':
				strRet += "&gt;";
				break;
			case '\'':
				strRet += "&apos;";
				break;
			case '"':
				strRet += "&quot;";
				break;
			default:
				ASSERT ("*pch must in ['&' '<' '>' '\'' '\"']" && false);
				break;
			}
		}
		else
		{
			if (i == XMLENCODE_BUFFSIZE)
			{
				pszEncodedText[i] = 0;
				strRet += pszEncodedText;
				i = 0;
			}

			pszEncodedText[i++] = *pch;
		}

		pch++;
	}
	
	if (i > 0)
	{
		pszEncodedText[i] = 0;
		strRet += pszEncodedText;
	}

	ASSERT (DecodeXML(strRet) == szSrc);

	return strRet;
}

string CIkXMLDom::DecodeXML(const char * szSrc, int len)
{
	if (szSrc == NULL)
		return "";

	if (len == -1)
		len = strlen(szSrc);

	string strRet;


	const char *pEnd = szSrc + len;
	const char *pch = szSrc;
	
	do
	{
		const char *pNext = FIND_CHR(pch, '&', pEnd);
		if (pNext)
		{
			STR_APPEND(strRet, pch, pNext - pch);
			
			pch = pNext;

			pNext = FIND_CHR(pch, ';', pEnd);
			if (pNext && (pNext-pch) < 6)
			{
				pNext++;

				string strTok;

				STR_ASSIGN(strTok, pch, pNext - pch );

				pch = pNext;

				if (strTok == "&amp;")
					strRet += '&';
				else if (strTok == "&lt;")
					strRet += '<';
				else if (strTok == "&gt;")
					strRet += '>';
				else if (strTok == "&apos;")
					strRet += '\'';
				else if (strTok == "&quot;")
					strRet += '"';
				else
					strRet += strTok;
			}
			else
			{
//				TRACE ("[%s] is wrong encoded xml\n", pch);
				//strRet += *pch++;
				return "";
			}
		}
		else
		{
			//strRet.append(pch, pEnd - pch);			
			STR_APPEND(strRet, pch, pEnd - pch );

			pch = pEnd;
			break;
		}
	}
	while (pch < pEnd);

//	ASSERT (EncodeXML(strRet) == szSrc);

	return strRet;
}

////////////////////////////////////////////////////////////////////////////////////////
// CIkXMLNode

CIkXMLDom::CIkXMLNode::CIkXMLNode() : m_pFirstChild(NULL),
	m_pLastChild(NULL),
	m_pPrevSibling(NULL),
	m_pNextSibling(NULL),
	m_pParentNode(NULL),
	m_pFirstAttribute(NULL),
	m_bNoTailTag(false),
	m_chQuotation('"'),
	m_bIsComment(false)
{
}

CIkXMLDom::CIkXMLNode::~CIkXMLNode()
{
	Destroy();
}

void CIkXMLDom::CIkXMLNode::Destroy()
{
	CIkXMLAttribute *pAttr = m_pFirstAttribute;
	CIkXMLAttribute *pAttrThis;
	while (pAttr)
	{
		pAttrThis = pAttr;
		pAttr = pAttr->m_pNext;
		delete pAttrThis;
	}

	m_pFirstAttribute = NULL;
	
	CIkXMLNode * pNode = m_pFirstChild;
	CIkXMLNode * pNodeThis;
	while (pNode)
	{
		pNodeThis = pNode;

		pNode = pNode->m_pNextSibling;

		delete pNodeThis;
	}

	m_pFirstChild = m_pLastChild = NULL;
	m_pNextSibling = NULL;

	
	list<CIkXMLNodeList*>::iterator it = m_listXMLNodeList.begin();
	while (it != m_listXMLNodeList.end())
	{
		CIkXMLNodeList * pNodeList = *it;

		delete pNodeList;

		it++;
	}
	m_listXMLNodeList.clear();

}

string CIkXMLDom::CIkXMLNode::Getxml()
{
	if (STR_ISEMPTY(m_strNodeName))
	{
		if (m_bIsComment)
			return "<!--" + m_strNodeText + "-->";

		return EncodeXML(m_strNodeText);
	}

	string strXml = "<" + m_strNodeName;
	
	CIkXMLAttribute * pAttr = m_pFirstAttribute;
	while (pAttr)
	{
		strXml += " " + pAttr->m_strName + "=" + m_chQuotation + EncodeXML(pAttr->m_strText) + m_chQuotation;

		pAttr = pAttr->m_pNext;
	}

	if (m_bNoTailTag)
	{
		strXml += "/>";
	}
	else
	{
		strXml += ">";
		
		CIkXMLNode * pChild = m_pFirstChild;
		while (pChild)
		{
			strXml += pChild->Getxml();
			
			pChild = pChild->m_pNextSibling;
		}
		
		strXml += "</" + m_strNodeName + ">";
	}

	return strXml;
}

string CIkXMLDom::CIkXMLNode::Gettext() 
{
	if (STR_ISEMPTY(m_strNodeName))
	{
		if (m_bIsComment)
			return "";

		return m_strNodeText;
	}

	string strText;

	CIkXMLNode * pChild = m_pFirstChild;
	while (pChild)
	{
		strText += pChild->Gettext();

		pChild = pChild->m_pNextSibling;
	}

	return strText;
}

void CIkXMLDom::CIkXMLNode::Puttext(string text)
{
	if (STR_ISEMPTY(m_strNodeName))
	{
		m_strNodeText = text;
		return;
	}

	if (m_pFirstChild && STR_ISEMPTY(m_pFirstChild->GetnodeName()))
	{
		m_pFirstChild->m_strNodeText = text;
	}
	else
	{
		CIkXMLNode * pNewChild = new CIkXMLNode;
		pNewChild->m_strNodeText = text;

		pNewChild->m_pNextSibling =	m_pFirstChild;
		m_pFirstChild = pNewChild;
	}

	CIkXMLNode * pChild = m_pFirstChild->m_pNextSibling;
	while (pChild)
	{
		if ( STR_ISEMPTY(pChild->GetnodeName()) )
			pChild->m_strNodeText = "";

		pChild = pChild->m_pNextSibling;
	}
}


IIkXMLAttribute * CIkXMLDom::CIkXMLNode::traverseAttribute(IIkXMLAttribute * pAttr)
{

	if(pAttr == NULL)
		return m_pFirstAttribute;
	else
		return pAttr->getNextItem();

}

IIkXMLAttribute * CIkXMLDom::CIkXMLNode::getAttribute(const char * szAttrName)
{
	IIkXMLAttribute * pAttr = m_pFirstAttribute;
	while (pAttr)
	{
		if (pAttr->GetnodeName() == szAttrName)
			return pAttr;

		pAttr = pAttr->getNextItem();
	}

	return NULL;
}


void CIkXMLDom::CIkXMLNode::AddChild(CIkXMLDom::CIkXMLNode * pNewNode)
{
	pNewNode->m_pParentNode = this;

	if (!m_pFirstChild)
	{
		m_pFirstChild = pNewNode;
		m_pLastChild = pNewNode;


		ASSERT (pNewNode->m_pPrevSibling == NULL);
	}
	else
	{
		ASSERT (NULL == m_pLastChild->m_pNextSibling);

		m_pLastChild->m_pNextSibling = pNewNode;
		pNewNode->m_pPrevSibling = m_pLastChild;
		
		m_pLastChild = pNewNode;
	}

	ASSERT (pNewNode->m_pNextSibling == NULL);
}

void CIkXMLDom::CIkXMLNode::AddAttribute(CIkXMLDom::CIkXMLNode::CIkXMLAttribute * pNewAttr)
{
	if (!m_pFirstAttribute)
		m_pFirstAttribute = pNewAttr;
	else
	{
		CIkXMLAttribute * pLast = m_pFirstAttribute;
		while (pLast->m_pNext)
		{
			pLast = pLast->m_pNext;
		}
		
		pLast->m_pNext = pNewAttr;
	}

	ASSERT (pNewAttr->m_pNext == NULL);
}

IIkXMLNode * CIkXMLDom::CIkXMLNode::selectSingleNode(const char * szNodeName)
{
	CIkXMLNode * pParentNode = this;

	string strNodeNameTok;
	for (int i=0 ; true ; i++)
	{
		strNodeNameTok = get_str_tok(szNodeName, "/", i);
		if (STR_LENGTH(strNodeNameTok) == 0)
			break;

		CIkXMLNode * pChildNode = pParentNode->m_pFirstChild;
		pParentNode = NULL;

		for ( ; pChildNode != NULL ; pChildNode = pChildNode->m_pNextSibling)
		{
			if (pChildNode->m_strNodeName == strNodeNameTok ||
				strNodeNameTok == "*")
			{
				pParentNode = pChildNode;
				break;
			}
		}

		if (!pParentNode)
		{
			/*
			if (i==0 && this->m_strNodeName == strNodeNameTok)
			{
				pParentNode = this;
			}
			else
			*/
				break;
		}
	}

	return pParentNode;
}

IIkXMLNodeList * CIkXMLDom::CIkXMLNode::selectNodes(const char * szNodeName)
{
	bool bAll = false;
	const int len = strlen(szNodeName);
	if (len == 0)
		return NULL;

	IIkXMLNode * pFirstSelectedChild = selectSingleNode(szNodeName);

	string strNodeName_last_tok;

	////tskim boundchecker
	if ( szNodeName[len-1] == '*' && 
		(len-2 < 0 || szNodeName[len-2] == '/' || szNodeName[len-2] == '\0') )
		bAll = true;
	else
	{
		const char *p = szNodeName + len - 1;
		while (p > szNodeName)
		{
			if (*p == '/')
			{
				strNodeName_last_tok = p+1;
				break;
			}
			p--;
		}

		if (STR_ISEMPTY(strNodeName_last_tok))
			strNodeName_last_tok = szNodeName;
	}

	CIkXMLNodeList * pNodeList = new CIkXMLNodeList;

	IIkXMLNode * pChild = pFirstSelectedChild;
	while (pChild)
	{
		if (bAll || pChild->GetnodeName() == strNodeName_last_tok )
		{
			pNodeList->m_vecXMLNode.push_back( pChild );
		}

		pChild = pChild->GetnextSibling();
	}

	m_listXMLNodeList.push_back( pNodeList );

	return pNodeList;
}

IIkXMLNode * CIkXMLDom::CIkXMLNode::createChild(const char * szNodeName, const char * szText)
{
	CIkXMLNode * pNewNode = new CIkXMLNode;
	pNewNode->m_strNodeName = szNodeName;
	pNewNode->m_strNodeText = szText;

	AddChild( pNewNode );

	return pNewNode;
}

IIkXMLAttribute * CIkXMLDom::CIkXMLNode::createAttribute(const char * szAttrName, const char * szText)
{
	CIkXMLAttribute * pAttr = new CIkXMLAttribute;
	pAttr->m_strName = szAttrName;
	pAttr->m_strText = szText;

	AddAttribute( pAttr );

	return pAttr;
}

////////////////////////////////////////////////////////////////////////////////////////
// CIkXMLDom

CIkXMLDom::CIkXMLDom() : m_pTopNode(NULL), m_bPreserveWhiteSpace(false)
{
}

CIkXMLDom::~CIkXMLDom()
{
	Destroy();
}

void CIkXMLDom::Destroy()
{
	if (!m_pTopNode)
		return;

	CIkXMLNode * pNode = m_pTopNode;
	CIkXMLNode * pNodeThis;
	while (pNode)
	{
		pNodeThis = pNode;

		pNode = pNode->m_pNextSibling;

		delete pNodeThis;
	}
	m_pTopNode = NULL;
}

//static 
LONG CIkXMLDom::JumpXMLHeader(const char * szXml, const char * pEnd)
{

	const char * lpStr;
	//skip <?xml .... ?>

	lpStr = szXml;

	if( lpStr[0] != '<' )
		return -1;
	
	while( lpStr[1] == '?' )
	{		
		lpStr = FIND_CHR( lpStr+2, '<', pEnd );
		if( lpStr == NULL )
			return -1;
	}

	return (LONG)(lpStr-szXml);
}


bool CIkXMLDom::loadXML(const char * szXml)
{
	return loadXML( szXml, strlen(szXml) );
}

bool CIkXMLDom::loadXML(const char * szXml, int len)
{
	LONG st_id;

	Destroy();

	if( szXml == NULL )
		return false;

	ASSERT (NULL == m_pTopNode);

	if(-1 == len)
		len = strlen(szXml);

	if ( len < 5 )
		return false;

	const char *pDataEnd = szXml + len;

	if( !SKIP_WSPACE( szXml, pDataEnd ) )
		return false;		// data end  о,

	st_id = JumpXMLHeader(szXml, szXml+len);

	if (st_id < 0 )
		return false;

	const char *pRetNodeEnd = NULL;
	m_pTopNode = parseSingleNodeXml(szXml+st_id, pDataEnd, pRetNodeEnd );
	
	return m_pTopNode != NULL;
}


/* GARA_XML_GRAMMER 

    cQU := '
	cDQ := "
	cEQ := =
	cLT := <
	cGT := >
	cSL := /

	DOC	:= [NODEENTRY]|[COMMENT] EOF

    COMMENT := <!-- ..... -->

	NODEENTRY := STAG STRING|NODEENTRY ETAG

	STAG := < STRING ATTRIBUTES > 

	ATTRIBUTES := [ATTRIBUTE]

	ATTRIBUTE  := STRING cEQ QUOTED

	ETAG := < SL STRING >


	QUOTED := (cQU STRING cQU)|(cDQ STRING cDQ)

	STRING := [a...z]|[A...Z]|[0...9]
	

*/


bool CIkXMLDom::CIkXMLNode::parseTagHeaderXml(const char * szTagHeaderXml, const char * pEnd, string & strErrorMessage, bool bPreserveWhiteSpace)
{
	const char * pNext;

	if ( strncmp(szTagHeaderXml, "!--", 3) == 0 )
	{
		m_bIsComment = true;
		STR_ASSIGN(m_strNodeText, szTagHeaderXml+3, pEnd-szTagHeaderXml - 5);
		return true;
	}
	
	pNext = FIND_WSPACE(szTagHeaderXml, pEnd);
	if (!pNext)
	{
		STR_ASSIGN(m_strNodeName, szTagHeaderXml, pEnd-szTagHeaderXml);

		return true;
	}

	STR_ASSIGN(m_strNodeName, szTagHeaderXml, pNext - szTagHeaderXml);

	pNext++;

	while (pNext < pEnd)
	{
		SKIP_WSPACE (pNext, pEnd);
		if (pNext == pEnd)
			return true;

		const char *pTokEnd = FIND_CHR(pNext, '=', pEnd);
		if (!pTokEnd)
		{
			STR_ASSIGN(strErrorMessage, pNext, pEnd-pNext);
			strErrorMessage += " Attribute xml  '='  ã  ϴ.";
			
			return false;
		}

		CIkXMLAttribute *pAttr = new CIkXMLAttribute;
		STR_ASSIGN(pAttr->m_strName, pNext, pTokEnd-pNext);
		STR_RTRIM( pAttr->m_strName );		// 2003-08-21 翵 <xml aaa = "Ǫ"/>  ϱ 

		pNext = pTokEnd + 1;

		SKIP_WSPACE(pNext, pEnd);

		m_chQuotation = *pNext;
		
		if (m_chQuotation != '\"' &&
			m_chQuotation != '\'')
		{
			STR_ASSIGN(strErrorMessage, pNext, pEnd-pNext);
			strErrorMessage += " Attribute xml  ['] Ǵ [\"]  ã  ϴ.";
			
			delete pAttr;
			return false;
		}
		
		pNext++;

		pTokEnd = FIND_CHR(pNext, m_chQuotation, pEnd);
		if (!pTokEnd)
		{	
			STR_ASSIGN(strErrorMessage, pNext, pEnd-pNext);
			strErrorMessage += string(" Attribute xml   [") + m_chQuotation + "]  ã  ϴ.";
			
			delete pAttr;
			return false;
		}
		
		if (!bPreserveWhiteSpace)
			SKIP_WSPACE(pNext, pTokEnd);
		
		STR_ASSIGN(pAttr->m_strText, pNext, pTokEnd-pNext);
		
		if (!bPreserveWhiteSpace)
			STR_RTRIM(pAttr->m_strText);
		
		string strDecodedText = DecodeXML( pAttr->m_strText );
		if( !STR_ISEMPTY(pAttr->m_strText) && STR_ISEMPTY(strDecodedText) )
		{
			delete pAttr;
			return false;
		}
		pAttr->m_strText = strDecodedText;
		
		pNext = pTokEnd+1;

		AddAttribute( pAttr );
	}

	return true;
}


CIkXMLDom::CIkXMLNode * CIkXMLDom::parseSingleNodeXml(const char *szXml, const char *pDataEnd, const char * &pRetNodeEnd)
{

	const char * p = szXml;
	const char * pNext;


	if (*p != '<')
	{
		m_strLastErrorMessage = "'<'  xml  ۵ ʾҽϴ.";
		return NULL;
	}

	p++;

	if (strncmp(p, "!--", 3) == 0)
	{
		// ּΰ..
		pNext = FIND_STR(p, "-->", pDataEnd);
		
		if (!pNext)
		{
			m_strLastErrorMessage = "'<!--'   '-->' ã  ϴ.";
			return NULL;
		}

		pNext += 2;
	}
	else
	{
		pNext = FIND_CHR(p, '>', pDataEnd);
		
		if (!pNext)
		{
			m_strLastErrorMessage = "'<'   '>' ã  ϴ.";
			return NULL;
		}
		else
		{
			if (FIND_CHR(p, '<', pDataEnd) < pNext)
			{
				m_strLastErrorMessage = "'<'   '>'   '<' Խϴ.";
				return NULL;
			}
		}
	}


	if (*(pNext-1) == '/')
		pNext--;

	CIkXMLNode * pNode = new CIkXMLNode;

	if (!pNode->parseTagHeaderXml( p, pNext, m_strLastErrorMessage, m_bPreserveWhiteSpace ))
	{
		delete pNode;
		return NULL;
	}

	if (pNode->IsComment())
	{
		pNode->m_bNoTailTag = true;
		pRetNodeEnd = pNext + 1;

		return pNode;
	}

	if (*pNext == '/')
	{
		pNode->m_bNoTailTag = true;
		pRetNodeEnd = pNext + 2;

		return pNode;
	}

	//pNode->m_bNoTailTag = false;
	ASSERT (pNode->m_bNoTailTag == false);

	p = pNext + 1;

	while (*p)
	{
		if (*p != '<')
		{
			pNext = FIND_CHR(p+1, '<', pDataEnd);
			if (!pNext)
			{
				m_strLastErrorMessage = pNode->GetnodeName() + "  ۵ '<'  ã  ϴ.";

				// fail..
				delete pNode;
				return NULL;
			}
			else
			{
				CIkXMLNode * pChildNode = new CIkXMLNode;

				if (!m_bPreserveWhiteSpace)
					SKIP_WSPACE(p, pNext);

				STR_ASSIGN(pChildNode->m_strNodeText, p, pNext-p);

				if (!m_bPreserveWhiteSpace)
					STR_RTRIM(pChildNode->m_strNodeText);

				string strDecodedText = DecodeXML( pChildNode->m_strNodeText );
				if( pChildNode->m_strNodeText != "" && strDecodedText == "" )
				{	
					delete pChildNode;
					delete pNode;
					return NULL;
				}

				pChildNode->m_strNodeText = strDecodedText;

				pNode->AddChild(pChildNode);

				p = pNext;
			}
		}
		else
		{
			if (*(p+1) == '/')
			{
				int len = STR_LENGTH(pNode->GetnodeName());
				if (strncmp(p+2, STR_C_STR(pNode->GetnodeName()), len) == 0 )
				{
					pRetNodeEnd = p+2+len+1;
					return pNode;
				}
				else
				{
					m_strLastErrorMessage = "</" + pNode->GetnodeName() + ">  ã  ϴ.";

					delete pNode;
					return NULL;
				}
			}
			else
			{
				CIkXMLNode * pChildNode = parseSingleNodeXml(p, pDataEnd, pNext);
				
				if (!pChildNode)
				{
					delete pNode;
					return NULL;
				}
				
				pNode->AddChild( pChildNode );
				
				p = pNext;
			}
		}
	}

	return pNode;
}

